#include "hmbdc/Copyright.hpp"
#pragma once


#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>

#include <string>
#include <memory>
#include <list>
#include <vector>
#include <unordered_set>
#include <sstream>


/**
 * @namespace hmbdc::app
 * hmbdc's application layer where API resides
 */
namespace hmbdc { namespace app { 

namespace config_detail {

using namespace std;
using namespace boost::property_tree;
/**
 * @brief class to hold an hmbdc configuration
 * @details it is based on a two level (fallback and section) json. 
 * top level for the fallback values and lower level for the section specific values.
 * a Config instance is always constructed to be associated to 0 or 1 specific section.
 * shown below:
 * 
 *      {
 *          "parameter_1": "top level, value used as a fallback",
 *          "parameter_2": "fallback is used when not configured in a section",
 *          "section_A": {
 *               "parameter_2": "lower level, a specific value effective in section_A",
 *               "parameter_1": "a specific value effective in section_A"
 *          },
 *          "another_section": {
 *               "parameter_2": "a specific value effective in another_section"
 *          }
 *      }
 *  json array is not supported !!!
 */
struct Config 
: ptree {
    using ptree::ptree;
    using Base = boost::property_tree::ptree;

    /**
     * @brief empty config
     */
    Config(){}

    /**
     * @brief construct using stream, optionally specifying the section name
     * @details if the section is nullptr, just use the fallback values. if
     * the section name cannot be found, throw an exception
     * 
     * @param is stream as input providing a json stream
     * @param section pointing to the effective section in the json above
     */
    explicit 
    Config(istream&& is, char const* section = nullptr) {
        if (section) {
            read_json(is, fallback_);
            (ptree&)*this = fallback_.get_child(section);
        } else {
            read_json(is, (ptree&)*this);
        }
    }
    /**
     * @brief construct using stream, optionally specifying the section name
     * @details if the section is nullptr, just use the fallback values. if
     * the section name cannot be found, throw an exception
     * 
     * @param is stream as input providing a json stream
     * @param section pointing to the effective section in the json above
     */
    explicit 
    Config(istream& is, char const* section = nullptr) {
        if (section) {
            read_json(is, fallback_);
            (ptree&)*this = fallback_.get_child(section);
        } else {
            read_json(is, (ptree&)*this);
        }
    }

    /**
     * @brief construct using a string, optionally specifying the section name
     * @details if the section is nullptr, just use the fallback values. if
     * the section name cannot be found, throw an exception
     * 
     * @param json string as input providing a json text
     * @param section pointing to the effective section in the json above
     */
    explicit 
    Config(char const* json, char const* section = nullptr)
    : Config(istringstream(json), section)
    {}

    /**
     * @brief construct using another ptree as fallbacks, optionally specifying the section name
     * @details if the section is nullptr, just use the fallback values. if
     * the section name cannot be found, throw an exception
     * 
     * @param t ptree as input providing fallbacks (and sections)
     * @param section pointing to the effective section in the ptree above
     */
    Config(ptree const& t, char const* section = nullptr) {
        if (section) {
            fallback_ = t;
            (ptree&)*this = fallback_.get_child(section);
        } else {
            (ptree&)*this = t;
        }
    }

    /**
     * @brief construct using a fallback ptree, and specify the section ptree
     * @details if the section is nullptr, just use the fallback values. if
     * the section name cannot be found, throw an exception
     * 
     * @param t ptree as input providing fallbacks
     * @param section ptree providing section specific values
     */
    Config(ptree const& t, ptree const& section) {
        fallback_ = t;
        (ptree&)*this = section;
    }

    /**
     * @brief internal use
     * @details this is to set the default user config values (not the fallback values)
     * 
     * @param c a config holding configuration values
     */
    void setDefaultUserConfig(Config const& c) {
        defaultUserConfig_.reset(new Config(c));
    }

    /**
     * @brief      Gets the child from the config.
     * @details check the section for it, if not found, try use fallback provided
     * if still missing, search using the default user config values set by
     * setDefaultUserConfig. Throw exception ptree_bad_path if all fail
     *
     * @param[in]  param  config parameter name
     *
     * @return     ptree reference to the child.
     */
    ptree const& getChildExt(const path_type& param) {
        auto res = get_child_optional(param);
        if (!res) {
            res = fallback_.get_child_optional(param);
            if (!res) {
                if (defaultUserConfig_) {
                    return defaultUserConfig_->getChildExt(param);
                } else {
                    throw ptree_bad_path("invalid param and no default user Config set", param);
                }
            }
        }
        return *res;
    }

    /**
     * @brief get a value from the config
     * @details check the section for it, if not found, try use fallback provided
     * if still missing, search using the default user config values set by
     * setDefaultUserConfig. Throw exception ptree_bad_path if all fail
     * 
     * @param param config parameter name
     * @tparam T type of the value
     * @return result
     */
    template <typename T>
    T getExt(const path_type& param) const {
        auto res = get_optional<T>(param);
        if (!res) {
            res = fallback_.get_optional<T>(param);
            if (!res) {
                if (defaultUserConfig_) {
                    return defaultUserConfig_->getExt<T>(param);
                } else {
                    throw ptree_bad_path("invalid param and no default user Config set", param); 
                }
            }
        }
        return *res; 
    }

    /**
     * @brief get a number value in hex format
     * @details check the section for it, if not found, try use fallback provided
     * if still missing, search using the default user config values set by
     * setDefaultUserConfig. Throw exception ptree_bad_path if all fail
     * 
     * @param param config parameter name
     * @tparam T numeric type of the value: int , uint64_t ...
     * @return result
     */
    template <typename T>
    T getHex(ptree::path_type const& param) const {
        istringstream iss(getExt<string>(param));
        T res;
        iss >> hex >> res;
        return res;
    }
 
    /**
     * @brief fill in a variable with a configured value retrieved using getExt
     * @details example cfg(abc, "abc")(def, "def");
     * 
     * @param to destination
     * @param param config parameter
     * 
     * @return the Config object itself
     */
    template <typename T>
    Config const& operator()(T& to, const path_type& param) const {
        to = getExt<T>(param);
        return *this;
    }

    /**
     * @brief fill an unordered_set with a configured value retrieved using getExt
     * @details the value in the Config is a space separated string
     * 
     * @param to destination
     * @param param config parameter
     * 
     * @return the Config object itself
     */
    template <typename T>
    Config const& operator()(std::unordered_set<T>& to, const path_type& param) const {
        // auto toStr = getExt<string>(param);
        auto s = getExt<string>(param);
        istringstream iss(s);
   
        for (auto iit = istream_iterator<T>(iss)
            ; iit != istream_iterator<T>()
            ; iit ++) {
            to.insert(*iit);
        }
        if (!iss.eof()) {
            throw (ptree_bad_data("not space separated items in Config ", param));
        }
   
        return *this;
    }

    /**
     * @brief fill an list with a configured value retrieved using getExt
     * @details the value in the Config is a json array
     * 
     * @param to destination
     * @param param config parameter
     * 
     * @return the Config object itself
     */
    template <typename T>
    Config const& operator()(std::vector<T>& to, const path_type& param) const {
        // auto toStr = getExt<string>(param);
        auto s = getExt<string>(param);
        istringstream iss(s);
   
        for (auto iit = istream_iterator<T>(iss)
            ; iit != istream_iterator<T>()
            ; iit ++) {
            to.emplace_back(*iit);
        }
        if (!iss.eof()) {
            throw (ptree_bad_data("not space separated items in Config ", param));
        }
   
        return *this;
    }

    /**
     * @brief get contents of all  the effective configure in the form of list of string pairs
     * @details only effective ones are shown
     * 
     * @param skipThese skip those config params
     * @return list of string pairs in the original order of ptree nodes
     */
    list<pair<string, string>> content(
        unordered_set<string> const& skipThese = unordered_set<string>()) const {
         list<pair<string, string>> res;
         unordered_set<string> history(skipThese);
         for (auto& p : *this) {
            if (history.find(p.first) == history.end()) {
                history.insert(p.first);
                if (p.second.empty()) { //leaf
                    res.push_back(make_pair(p.first, p.second.get_value<string>()));
                }
                 // else { //or array
                 //    auto arrayText = getArrayText(p.second);
                 //    if (arrayText.size()) {
                 //        res.push_back(make_pair(p.first, arrayText));
                 //    }
                 // }
            }
         }
         for (auto& p : fallback_) {
            if (history.find(p.first) == history.end()) {
                history.insert(p.first);
                if (p.second.empty()) { //leaf
                    res.push_back(make_pair(p.first, p.second.get_value<string>()));
                } 
                // else { // or array
                //     auto arrayText = getArrayText(p.second);
                //     if (arrayText.size()) {
                //         res.push_back(make_pair(p.first, arrayText));
                //     }
                // }
            }
         }
         if (defaultUserConfig_) {
            auto more = defaultUserConfig_->content(history);
            res.insert(res.end(), more.begin(), more.end());
         }
         return res;
    }
    
    /**
     * @brief      stream out the effective settings
     *
     * @param      os    ostream
     * @param      cfg   The configuration
     *
     * @return     os
     */
    friend ostream& operator << (ostream& os, Config const& cfg) {
        for (auto& r : cfg.content()) {
            os << r.first << '=' << r.second << endl;
        }
        return os;
    }

private:
    // string getArrayText(ptree const& pt) const {
    //     bool isArray = true;
    //     string arrayText("[");
    //     for (auto& q : pt) {
    //         if (!q.second.empty() || q.first.size() != 0) {
    //             isArray = false;
    //             break;
    //         }
    //         arrayText += "\"" + q.second.get_value<string>() + "\",";
    //     }
    //     *arrayText.rbegin() = ']';
    //     if (isArray) {
    //         return arrayText;
    //     } else {
    //         return string();
    //     }
    // }

    ptree fallback_;
    std::shared_ptr<Config> defaultUserConfig_;
};
} //config_detail

using Config = config_detail::Config;
}}

