#include "hmbdc/Copyright.hpp"
#pragma once 


#define NETMAP_WITH_LIBS
#include <net/netmap_user.h>
#undef NETMAP_WITH_LIBS

#include "hmbdc/app/netmap/Messages.hpp"
#include "hmbdc/app/Base.hpp"
#include "hmbdc/pattern/MonoLockFreeBuffer.hpp"
#include "hmbdc/comm/eth/Misc.h"
#include "hmbdc/time/Rater.hpp"
#include "hmbdc/numeric/BitMath.hpp"

#include <boost/regex.hpp>
#include <memory>
#include <type_traits>


#include <netinet/ether.h>      /* ether_aton */
#include <linux/if_packet.h>    /* sockaddr_ll */
#include <sys/sysctl.h> /* sysctl */
#include <ifaddrs.h>    /* getifaddrs */

#include <poll.h>    

namespace hmbdc { namespace app { namespace netmap { 

struct NetContext;
struct Sender;

namespace sendtransportengine_detail {
using namespace std;    
using namespace hmbdc::time;
using namespace hmbdc::comm::eth;

/**
 * @brief power a netmap port sending functions
 * @details this needs to be created using NetContext and start in an app::Context
 * 
 */
struct SendTransportEngine 
: Client<SendTransportEngine> {
    using ptr = std::shared_ptr<SendTransportEngine>;

private:
    uint16_t
    outBufferSizePower2();

    friend struct hmbdc::app::netmap::NetContext; //only be created by NetContext
    SendTransportEngine(Config const&, size_t); 
public:
    ~SendTransportEngine();
    bool droppedCb() override;

    /*virtual*/ 
    void invokedCb(uint16_t threadSerialNumber) HMBDC_RESTRICT override  {
        for (int i = nmd_->first_tx_ring; i <= nmd_->last_tx_ring; i++) {
            struct netmap_ring * txring = NETMAP_TXRING(nmd_->nifp, i);
            if (nm_ring_empty(txring))
                continue;

            sendPackets(txring);
        }
        if (hmbdc_unlikely(ioctl(nmd_->fd, NIOCTXSYNC, NULL) < 0)) {
            HMBDC_THROW(std::runtime_error, "IO error");
        }
    }

    void stoppedCb(std::exception const& e) override;

    char const* hmbdcName() const { 
        return this->hmbdcName_.c_str();
    }

    std::tuple<char const*, int> schedSpec() const {
        return std::make_tuple(this->schedPolicy_.c_str(), this->schedPriority_);
    }

private:
    bool match(Topic const& t) const {
        return boost::regex_match(t.c_str(), topicRegex_);
    }
    friend struct hmbdc::app::netmap::Sender; 
    template <typename... Messages>
    void queue(Topic const& t, Messages&&... msgs) HMBDC_RESTRICT {
        auto n = sizeof...(msgs);
        auto it = buffer_.claim(n);
        queue(it, t, std::forward<Messages>(msgs)...);
        buffer_.commit(it, n);
    }

    template <typename... Messages>
    bool tryQueue(Topic const& t, Messages&&... msgs) HMBDC_RESTRICT {
        auto n = sizeof...(msgs);
        auto it = buffer_.tryClaim(n);
        if (it) {
            queue(it, t, std::forward<Messages>(msgs)...);
            buffer_.commit(it, n);
            return true;
        }
        return false;
    }

    template <typename M, typename... Messages>
    void queue(pattern::MonoLockFreeBuffer::iterator it
        , Topic const& t, M&& msg, Messages&&... msgs) {
        using Message = typename std::remove_reference<M>::type;
        static_assert(std::is_trivially_destructible<Message>::value, "cannot send message with dtor");
        if (hmbdc_unlikely(sizeof(Message) > maxMessageSize_)) {
            HMBDC_THROW(std::out_of_range, "maxMessageSize too small to hold a message when constructing SendTransportEngine");        
        }
        auto s = *it;
        TransportMessageHeader::copyTo(s, t, std::forward<M>(msg));
        queue(++it, t, std::forward<M>(msgs)...);
    }

    template <typename Message, typename ... Args>
    void queueInPlace(Topic const& t, Args&&... args) HMBDC_RESTRICT {
        static_assert(std::is_trivially_destructible<Message>::value, "cannot send message with dtor");
        if (hmbdc_unlikely(sizeof(Message) > maxMessageSize_)) {
            HMBDC_THROW(std::out_of_range, "maxMessageSize too small to hold a message when constructing SendTransportEngine");        
        }
        auto s = buffer_.claim();
        TransportMessageHeader::copyToInPlace<Message>(*s, t, std::forward<Args>(args)...);
        buffer_.commit(s);
    }

    void queueBytes(Topic const& t, uint16_t tag, void const* bytes, size_t len);
    void queue(pattern::MonoLockFreeBuffer::iterator it
        , Topic const& t) {}

    void sendPackets(struct netmap_ring *);

    void getMacAddresses();
    static
    void initializePacket(struct pkt *, int, std::string, std::string, ether_addr, ether_addr, uint16_t, uint16_t);

    static 
    void updatePacket(struct pkt *, size_t, bool = true);

    Config const config_;
    string hmbdcName_;
    string schedPolicy_;
    int schedPriority_;
    std::string topic_;
    boost::regex topicRegex_;
    size_t maxMessageSize_;


    struct nm_desc *nmd_;
    pattern::MonoLockFreeBuffer buffer_;
    int virtHeader_; //v hdr len


    ether_addr srcEthAddr_;
    ether_addr dstEthAddr_;
    pkt precalculatedPacketHead_;
    bool doChecksum_;
    Rater rater_;
    size_t maxSendBatch_;
    uint16_t mtu_;
};

} //sendtransportengine_detail

using SendTransportEngine = sendtransportengine_detail::SendTransportEngine;

}}}
