#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Logger.hpp"
#include "hmbdc/app/tcpcast/Transport.hpp"
#include "hmbdc/app/tcpcast/Messages.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/os/DownloadFile.hpp"
#include "hmbdc/os/DownloadMemory.hpp"
#include "hmbdc/comm/inet/Misc.hpp"

#include "hmbdc/app/tcpcast/RecvSession.hpp"

#include <boost/lexical_cast.hpp>

#include <functional>
#include <memory>
#include <utility>

#include <iostream>

namespace hmbdc { namespace app { namespace tcpcast {

namespace recvsession_detail {
using namespace std;
using namespace hmbdc::text;

template <typename OutputBuffer, typename MsgArbitrator>
struct RecvSession {
    using ptr = shared_ptr<RecvSession<OutputBuffer, MsgArbitrator>>;
    using CleanupFunc = std::function<void()>;
    RecvSession(Config const& config
        , StringTrieSet const& subscriptions
        , boost::regex const& topicRegex
        , OutputBuffer& outputBuffer
        , uint64_t connKey
        , MsgArbitrator& arb)
    : config_(config)
    , subscriptions_(subscriptions)
    , topicRegex_(topicRegex)
    , writeFd_(config)
    , arb_(arb)
    , outputBuffer_(outputBuffer)
    , connKey_(connKey)
    , initialized_(false)
    , stopped_(false)
    , bufSize_(config.getExt<size_t>("maxTcpReadBytes"))
    , buf_((char*)memalign(SMP_CACHE_BYTES, bufSize_))
    , bufCur_(buf_)
    , filledLen_(0)
    , currTransportHeadFlag_(0) {
        config(allowRecvMemoryAttachment_, "allowRecvMemoryAttachment");
    }

    ~RecvSession() {
        free(buf_);
        HMBDC_LOG_N("RecvSession retired: ", id());
    }

    void start(in_addr_t ip, uint16_t port) {
        sockaddr_in remoteAddr = {0};
        remoteAddr.sin_family = AF_INET;
        remoteAddr.sin_addr.s_addr = ip;
        remoteAddr.sin_port = htons(port);
        if (connect(writeFd_.fd, (sockaddr*)&remoteAddr, sizeof(remoteAddr)) < 0) {
            if (errno != EINPROGRESS) {
                HMBDC_THROW(runtime_error, "connect fail, errno=" << errno);
            }
        }

        auto forRead = dup(writeFd_.fd);
        if (forRead == -1) {
            HMBDC_THROW(std::runtime_error, "dup failed errno=" << errno);
        }

        readFd_.fd = forRead;
        utils::EpollTask::instance().add(EPOLLOUT|EPOLLET, writeFd_);
        utils::EpollTask::instance().add(EPOLLIN|EPOLLET, readFd_);
    }

    void stop() {
        if (currTransportHeadFlag_ == hasMemoryAttachment::flag) {
            new (*itPending_) MessageWrap<Flush>;
            outputBuffer_.commit(itPending_);
            currTransportHeadFlag_ = 0;
        }
        outputBuffer_.putSome(sessionDropped_);
    }

    void sendSubscribe(Topic const& t) {
        if (hmbdc_unlikely(boost::regex_match(t.c_str(), topicRegex_))) {
            char buf[70];
            buf[0]='+';
            auto l = int(t.copyTo(buf + 1));
            buf[l + 1] = '\t';
            if (hmbdc_unlikely(send(writeFd_.fd, buf, l + 2, MSG_NOSIGNAL) != l + 2)) {
                HMBDC_LOG_C("error when sending subscribe to ", id());
                stopped_ = true;
            }
        }
    }


    void sendUnsubscribe(Topic const& t) {
        if (hmbdc_unlikely(boost::regex_match(t.c_str(), topicRegex_))) {
            char buf[70];
            buf[0]='-';
            auto l = int(t.copyTo(buf + 1));
            buf[l + 1] = '\t';
            if (hmbdc_unlikely(l + 2 != send(writeFd_.fd, buf, l + 2, MSG_NOSIGNAL))) {
                HMBDC_LOG_C("error when sending unsubscribe to ", id(), " errno=", errno);
                stopped_ = true;
            }
        }
    }

    void heartbeat() {
        if (!initialized_) return;
        char const* hb = "+\t"; 
        if (hmbdc_unlikely(2 != send(writeFd_.fd, hb, 2, MSG_NOSIGNAL))){
            HMBDC_LOG_C("error when sending heartbeat to ", id(), " errno=", errno);
            stopped_ = true;
        }
    }

    char const* id() const {
        return id_.c_str();
    }

    bool runOnce() {
        if (hmbdc_unlikely(stopped_)) return false;
        if (hmbdc_unlikely(!initialized_ && writeFd_.isFdReady())) {
            try {
                initializeConn();
            } catch (std::exception const& e) {
                HMBDC_LOG_W(e.what());
                return false;
            } catch (...) {
                HMBDC_LOG_C("unknown exception");
                return false;
            }
        }
        return doRead();
    }

private:
    void initializeConn() {
        int flags = fcntl(writeFd_.fd, F_GETFL, 0);
        flags &= ~O_NONBLOCK;
        if (fcntl(writeFd_.fd, F_SETFL, flags) < 0) {
            HMBDC_THROW(std::runtime_error, "fcntl failed errno=" << errno);
        }

        auto sz = config_.getExt<int>("tcpRecvBufferBytes");
        if (sz) {
            if (setsockopt(readFd_.fd, SOL_SOCKET, SO_RCVBUF, &sz, sizeof(sz)) < 0) {
                HMBDC_LOG_C("failed to set send buffer size=", sz);
            }
        }
        
        auto addrPort = hmbdc::comm::inet::getPeerIpPort(writeFd_.fd);
        id_ = addrPort.first + ":" + std::to_string(addrPort.second);
        strncpy(sessionStarted_.payload.ip, id_.c_str()
            , sizeof(sessionStarted_.payload.ip));
        sessionStarted_.payload.ip[sizeof(sessionStarted_.payload.ip) - 1] = 0;
        strncpy(sessionDropped_.payload.ip, id_.c_str()
            , sizeof(sessionDropped_.payload.ip));
        sessionDropped_.payload.ip[sizeof(sessionDropped_.payload.ip) - 1] = 0;
        outputBuffer_.putSome(sessionStarted_);
        HMBDC_LOG_N("RecvSession started: ", id());
        sendSubscriptions();
        initialized_ = true;
    }

    void sendSubscriptions() {
        ostringstream oss;
        oss << '@' << connKey_ << '\t';
        for (auto it = subscriptions_.begin(); it != subscriptions_.end(); ++it) {
            auto const& t = it->first;
            if (boost::regex_match(t.c_str(), topicRegex_)) {
                oss << '+' << t;
                if (it->second) {
                    oss << '*';
                }
                oss << '\t';
            }
        }
        auto s = oss.str() + "+\t"; //mark all sent with "+\t"
        auto sz = size_t(send(writeFd_.fd, s.c_str(), s.size(), MSG_NOSIGNAL));
        if (hmbdc_unlikely(sz != s.size())) {
            HMBDC_LOG_C("error when sending subscriptions to ", id());
            stopped_ = true;
        }
    }

    bool doRead() {
        do {
            if (hmbdc_unlikely(currTransportHeadFlag_ == hasMemoryAttachment::flag
                && filledLen_)) {
                auto s = memoryAttachment_.write(bufCur_, filledLen_);
                if (memoryAttachment_.writeDone()) {
                    memoryAttachment_.close();
                    currTransportHeadFlag_ = 0;
                    outputBuffer_.commit(itPending_);
                }
                bufCur_ += s;
                filledLen_ -= s;
            }
            while (currTransportHeadFlag_ == 0 
                && filledLen_ >= sizeof(TransportMessageHeader)) {
                auto h = reinterpret_cast<TransportMessageHeader*>(bufCur_);
                auto wireSize = h->wireSize();
                if (hmbdc_likely(filledLen_ >= wireSize)) {
                    if (hmbdc_likely(subscriptions_.check(h->topic()))) {
                        auto a = arb_(h);
                        if (hmbdc_likely(a > 0)) {
                            if (hmbdc_unlikely(h->flag == hasMemoryAttachment::flag
                                && allowRecvMemoryAttachment_.find(h->typeTag()) != allowRecvMemoryAttachment_.end())) {
                                currTransportHeadFlag_ = h->flag;
                                itPending_ = outputBuffer_.claim();
                                char* b = static_cast<char*>(*itPending_);
                                auto l = std::min<size_t>(outputBuffer_.maxItemSize(), h->messagePayloadLen);
                                memcpy(b, h->payload(), l);
                                auto& att = 
                                    static_cast<app::MessageWrap<hasMemoryAttachment>*>(*itPending_)->payload;
                                att.hasMemoryAttachment::attachment = memoryAttachment_.open(att.len);
                                att.hasMemoryAttachment::afterConsumedCleanupFunc = hasMemoryAttachment::free;
                            } else if (h->flag != hasMemoryAttachment::flag) {
                                auto l = std::min<size_t>(outputBuffer_.maxItemSize(), h->messagePayloadLen);
                                outputBuffer_.put(h->payload(), l);
                                currTransportHeadFlag_ = 0;
                            } //else ignore
                        } else if (hmbdc_unlikely(a == 0)) {
                            //cannot decide now - keep bufCur_ filledLen_ unchanged 
                            //try it later, use a timer since there might be no more bytes arriving
                            return true;
                        } //else move to next msg
                    }
                    bufCur_ += wireSize;
                    //memmove(buf_, buf_ + h->wireSize(), filledLen_ - h->wireSize());
                    filledLen_ -= wireSize;
                } else {
                    break;
                }
            }
            memmove(buf_, bufCur_, filledLen_);
            bufCur_ = buf_;

            if (readFd_.isFdReady()) {
                auto l = recv(readFd_.fd, buf_ + filledLen_, bufSize_ - filledLen_
                    , MSG_NOSIGNAL|MSG_DONTWAIT);
                if (hmbdc_unlikely(l < 0)) {
                    if (hmbdc_unlikely(!readFd_.checkErr())) {
                        HMBDC_LOG_C("recv failed errno=", errno);
                        return false;
                    }
                    return true;
                } else if (hmbdc_unlikely(l == 0)) {
                    HMBDC_LOG_W("peer dropped:", id());
                    return false;
                }
                filledLen_ += l;
            } else {
                break;
            }
        } while (filledLen_ >= sizeof(TransportMessageHeader));
        return true;
    }

    Config const& config_;
    StringTrieSet const& subscriptions_;
    boost::regex topicRegex_;
    utils::EpollFd readFd_;
    EpollFd writeFd_;
    MsgArbitrator& arb_;
    OutputBuffer& outputBuffer_;
    uint64_t connKey_;
    string id_;
    MessageWrap<SessionStarted> sessionStarted_;
    MessageWrap<SessionDropped> sessionDropped_;
    bool initialized_;
    bool stopped_;
    size_t const bufSize_;
    char* buf_;
    char* bufCur_;
    size_t filledLen_;
    uint8_t currTransportHeadFlag_;
    typename OutputBuffer::iterator itPending_;
    
    std::unordered_set<uint16_t> allowRecvMemoryAttachment_;
    os::DownloadMemory memoryAttachment_;
};
} //recvsession_detail
template <typename OutputBuffer, typename MsgArbitrator>
using RecvSession = recvsession_detail::RecvSession<OutputBuffer, MsgArbitrator>;
}}}
