#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/tcpcast/Transport.hpp"
#include "hmbdc/app/tcpcast/RecvSession.hpp"
#include "hmbdc/app/tcpcast/Messages.hpp"
#include "hmbdc/app/udpcast/RecvTransportEngine.hpp"
#include "hmbdc/app/Logger.hpp"
#include "hmbdc/text/StringTrieSet.hpp"

#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/unordered_map.hpp>

#include <memory>

namespace hmbdc { namespace app { namespace tcpcast {

/**
 * @brief interface to power a tcpcast transport receiving functions
 */
struct RecvTransport 
: Transport {
    using ptr = std::shared_ptr<RecvTransport>;
    using Transport::Transport;
    virtual ~RecvTransport(){}
    /**
     * @brief a take all arbitrator (no arbitration at all)
     * @details it provides the default type for arbitration which does nothing
     * it also provides a template on writing a user defined arbitrator
     * 
     * @param  h handle to retrieve what is inside of the message
     * @return always returns 1 here
     * In general: 1 keep the message; -1 to drop; 
     * 0 cannot decide, ask me later.
     */
    struct NoOpArb {
        int operator()(TransportMessageHeader const* h) {
            return 1; //always keep it
        }
    };

private:
    friend struct NetContext; //only be used by NetContext
    virtual void listenTo(comm::Topic const& t) = 0;
    virtual void stopListenTo(comm::Topic const& t) = 0;
};


namespace recvtransportengine_detail {
using namespace hmbdc::time;
using namespace std;
/**
 * @brief impl class
 * @details this needs to be created using NetContext and start in an app::Context
 * 
 * @tparam OutputBuffer type of buffer to hold resulting network messages
 */
template <typename OutputBuffer, typename MsgArbitrator>
struct RecvTransportEngine 
: RecvTransport
, TimerManager
, ReoccuringTimer
, Client<RecvTransportEngine<OutputBuffer, MsgArbitrator>>
, MessageHandler<RecvTransportEngine<OutputBuffer, MsgArbitrator>
    , Subscribe, Unsubscribe, TopicSource> {
    using SELF = RecvTransportEngine;
    using MH = MessageHandler<SELF, Subscribe, Unsubscribe, TopicSource>;
    friend struct NetContext; //only be created by NetContext

/**
 * @brief ctor
 * @details io_service could be passed in by user, in this case NO more than two threads should
 * power this io_service instance since that would violate the thread garantee of Client, which
 * is no callbacks are called in parallel
 * 
 * @param cfg specify the details of the tcpcast transport
 * @param outputBuffer holding the results
 * @param arb arbitrator instance to decide which messages to drop and keep; it ONLY supports 
 * hmbdc message (AFTER topic filtering) level (NO packet level since it is tcp)
 */
    RecvTransportEngine(Config const& cfg
        , OutputBuffer& outputBuffer
        , MsgArbitrator arb = NoOpArb())
    : RecvTransport(cfg)
    , ReoccuringTimer(Duration::seconds(cfg.getExt<size_t>("heartbeatPeriodSeconds")))
    , config_(cfg)
    , mcConfig_(cfg) 
    , buffer_(std::max({sizeof(MessageWrap<Subscribe>)
        , sizeof(MessageWrap<Unsubscribe>)
        , sizeof(MessageWrap<TopicSource>)
        }), config_.getExt<uint16_t>("cmdBufferSizePower2"))
    , outputBuffer_(outputBuffer)
    , maxItemSize_(outputBuffer.maxItemSize())
    , arb_(arb) 
    , myIp_(inet_addr(hmbdc::comm::inet::getLocalIpMatchMask(
        cfg.getExt<string>("ifaceAddr")).c_str()))
    , mcastEmuProxyScanTimer_(
        Duration::seconds(cfg.getExt<size_t>("mcastEmuProxyScanPeriodSeconds"))) {
        if (outputBuffer.maxItemSize() < sizeof(SessionStarted) 
            || outputBuffer.maxItemSize() < sizeof(SessionDropped))  {
            HMBDC_THROW(std::out_of_range
                , "buffer is too small for notification: SessionStarted and SessionDropped");
        }
        mcConfig_.put("outBufferSizePower2", 3u);
        mcReceiver_.reset(new udpcast::RecvTransportImpl<decltype(buffer_)>(
            mcConfig_, buffer_, udpcast::RecvTransport::NoOpArb()));

        if (cfg.getExt<bool>("useTcpcastMcProxy")) {
            config_(tcpcastMcEmuProxyAddrs_, "udpcastDests");
            if (tcpcastMcEmuProxyAddrs_.size() == 0) {
                HMBDC_THROW(std::out_of_range, "no TcpcastEmuProxy specified");
            }
            topicSinkMsgBufLen_ = 
                udpcast::SendTransport::encode(tcpcastAdTopic_, topicSinkMsgBuf_, sizeof(topicSinkMsgBuf_)
                , TopicSink{cfg.getExt<std::string>("udpcastListenAddr")
                    , cfg.getExt<uint16_t>("udpcastListenPort")})->wireSize();
        }
    }

/**
 * @brief start the show by schedule the message recv
 */
    /*virtual*/ 
    void messageDispatchingStartedCb(uint16_t threadSerialNumber) override {
        mcReceiver_->start();
        mcReceiver_->listenTo(tcpcastAdTopic_);

        setCallback(
            [this](TimerManager& tm, SysTime const& now) {
                for (auto& s : recvSessions_) {
                    s.second->heartbeat();
                }
            }
        );

        schedule(SysTime::now(), *this);
        if (tcpcastMcEmuProxyAddrs_.size()) {
            mcastEmuProxyScanTimer_.setCallback(
                [this](TimerManager& tm, SysTime const& now) {
                    for (auto const& s : tcpcastMcEmuProxyAddrs_) {
                        sendto(mcReceiver_->fd, topicSinkMsgBuf_, topicSinkMsgBufLen_
                            , MSG_NOSIGNAL|MSG_DONTWAIT, (struct sockaddr *)&s.v, sizeof(s.v));
                    }
                }
            );
            schedule(SysTime::now(), mcastEmuProxyScanTimer_);
        }
    };

    using Transport::hmbdcName;
    using Transport::schedSpec;

/**
 * @brief should not happen ever unless an exception thrown
 * 
 * @param e exception thown
 */
    /*virtual*/ 
    void stoppedCb(std::exception const& e) override {
        HMBDC_LOG_C(e.what());
    };

/**
 * @brief power the io_service and other things
 * 
 */
    /*virtual*/ 
    void invokedCb(uint16_t threadSerialNumber) HMBDC_RESTRICT override  {
        for (auto it = recvSessions_.begin(); it != recvSessions_.end();) {
            if (hmbdc_unlikely(!(*it).second->runOnce())) {
                (*it).second->stop();
                recvSessions_.erase(it++);
            } else {
                it++;
            }
        }
        pattern::MonoLockFreeBuffer::iterator begin, end;
        auto n = buffer_.peek(begin, end);
        auto it = begin;
        while (it != end) {
            MH::handleMessage(*static_cast<MessageHead*>(*it++));
        }
        buffer_.wasteAfterPeek(begin, n);

        mcReceiver_->runOnce(true);
    }
   
/**
 * @brief only used by MH
 */
    void handleMessageCb(Subscribe const& t) {
        subscriptions_.add(boost::lexical_cast<std::string>(t.topic));
        for (auto& p : recvSessions_) {
            p.second->sendSubscribe(t.topic);
        }
    }

/**
 * @brief only used by MH
 */
    void handleMessageCb(Unsubscribe const& t) {
        subscriptions_.erase(boost::lexical_cast<std::string>(t.topic));
        for (auto& p : recvSessions_) {
            p.second->sendUnsubscribe(t.topic);
        }
    }

/**
 * @brief only used by MH
 */
    void handleMessageCb(TopicSource const& t) {
        auto ip = inet_addr(t.ip);
        //unless using loopback address, don't EVER listen to myself (own process)
        if (ip != 0x100007ful //127.0.0.1
            && ip == myIp_ 
            && (!t.loopback || t.pid == getpid())) {
            return;
        }
        auto key = make_pair(ip, t.port);
        if (recvSessions_.find(key) == recvSessions_.end()) {
            boost::regex topicRegex(t.topicRegex);
            for (auto it = subscriptions_.begin(); it != subscriptions_.end(); ++it) {
                if (boost::regex_match(it->first.c_str(), topicRegex)
                    || it->second) {
                    try {
                        auto sess = new Session(
                            config_
                            , subscriptions_
                            , topicRegex
                            , outputBuffer_
                            , t.connKey
                            , arb_
                        );
                        sess->start(ip, t.port);
                        recvSessions_[key] = typename Session::ptr(sess);
                    } catch (std::exception const& e) {
                        HMBDC_LOG_C(e.what());
                    }
                    break;
                }
            }
        } //else ignore
    }

private:
    using Session = RecvSession<OutputBuffer, MsgArbitrator>;
    void listenTo(Topic const& t) override {
        buffer_.put(MessageWrap<Subscribe>{t});
    }

    void stopListenTo(Topic const& t) override {
        buffer_.put(MessageWrap<Unsubscribe>{t});
    }
    Config config_, mcConfig_;
    pattern::MonoLockFreeBuffer buffer_;
    OutputBuffer& outputBuffer_;
    size_t maxItemSize_;
    unique_ptr<udpcast::RecvTransportImpl<decltype(buffer_)>> mcReceiver_;

    text::StringTrieSet subscriptions_;
    boost::unordered_map<pair<uint64_t, uint16_t>
        , typename Session::ptr> recvSessions_;
    MsgArbitrator arb_;
    uint32_t myIp_;
    std::vector<comm::inet::Endpoint> tcpcastMcEmuProxyAddrs_;
    ReoccuringTimer mcastEmuProxyScanTimer_;
    char topicSinkMsgBuf_[64];
    size_t topicSinkMsgBufLen_;
};

} //recvtransportengine_detail
template <typename OutputBuffer, typename MsgArbitrator>
using RecvTransportEngine = recvtransportengine_detail::RecvTransportEngine<OutputBuffer, MsgArbitrator>;
}}}
