#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Logger.hpp"
#include "hmbdc/app/tcpcast/Transport.hpp"
#include "hmbdc/app/tcpcast/Messages.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/pattern/MonoLockFreeBuffer.hpp"
#include "hmbdc/comm/inet/Misc.hpp"

#include "hmbdc/app/tcpcast/SendSession.hpp"

#include <unordered_set>
#include <memory>
#include <utility>
#include <iostream>

#include <netinet/tcp.h>

namespace hmbdc { namespace app { namespace tcpcast {

namespace sendserver_detail {
using namespace hmbdc::pattern;
using namespace std;

struct SendServer {
    SendServer(Config const& cfg
        , size_t toSendQueueMaxSize)
    : config_(cfg)
    , serverFd_(cfg)
    , serverAddr_(serverFd_.localAddr)
    , advertisingMessage_(
        cfg.getExt<string>("topicRegex")
        , serverFd_.localIp
        , serverFd_.localPort
        , config_.getExt<bool>("loopback")
    )
    , distribution_(1)
    , maxSendBatch_(config_.getExt<size_t>("maxSendBatch")) {
        if (listen(serverFd_.fd, 10) < 0) {
            HMBDC_THROW(runtime_error, "failed to listen, errno=" << errno);
        }
        HMBDC_LOG_N("listen at ", advertisingMessage_);
        utils::EpollTask::instance().add(EPOLLIN|EPOLLET, serverFd_);
        toSendQueue_.set_capacity(toSendQueueMaxSize);
    }

    TopicSource const& advertisingMessage() const {
        advertisingMessage_.connKey = distribution_(generator_);
        return advertisingMessage_;
    }

    void queue(pair<char const*, char const*> const& t
        , ToSend && toSend
        , size_t toSendByteSize
        , MonoLockFreeBuffer::iterator it) {
        toSendQueue_.push_back(forward_as_tuple(t, forward<ToSend>(toSend), toSendByteSize, it));
    }

    /**
     * @brief run the server's async send function and decide which items in buffer
     * can be release
     * @details called all the time
     * 
     * @param begin if nothing can be released in buffer return this
     * @param end if all can be releases return this
     * 
     * @return iterator in buffer poing to new start (not released)
     */
    MonoLockFreeBuffer::iterator runOnce(MonoLockFreeBuffer::iterator begin
        , MonoLockFreeBuffer::iterator end) HMBDC_RESTRICT {
        doAccept();
        if (sessions_.size() == 0) {
            toSendQueue_.clear();
            return end;
        }
        //nothing to retire by default
        auto newStartIt = begin;

        auto minIndex = toSendQueue_.size();
        for (auto it = sessions_.begin(); it != sessions_.end();) {
            if (minIndex > (*it)->toSendQueueIndex_) {
                minIndex = (*it)->toSendQueueIndex_;
            }
            if (hmbdc_unlikely(!(*it)->runOnce())) {
                sessions_.erase(it++);
            } else {
                it++;
            }
        }
        if (minIndex) {
            newStartIt = get<3>(toSendQueue_[minIndex - 1]);
            toSendQueue_.erase_begin(minIndex);
            for (auto it = sessions_.begin(); it != sessions_.end(); ++it) {
                (*it)->toSendQueueIndex_ -= minIndex;
            }
        }

        return newStartIt;
    }

    size_t readySessionCount() const {
        size_t res = 0;
        for (auto const& s : sessions_) {
            if (s->ready()) res++;
        }
        return res;
    }

    void killSlowestSession() {
        for (auto it = sessions_.begin(); it != sessions_.end(); ++it) {
            if (0 == (*it)->toSendQueueIndex_) {
                HMBDC_LOG_C((*it)->id(), " too slow, dropping");
                sessions_.erase(it);
                break;
            }
        }
    }

private:
    void doAccept() HMBDC_RESTRICT {
        if (hmbdc_unlikely(serverFd_.isFdReady())) {
            auto addrlen = sizeof(serverAddr_);
            auto conn = accept(serverFd_.fd, (struct sockaddr *)&serverAddr_, (socklen_t*)&addrlen);
            if (conn == -1) {
                if (!serverFd_.checkErr()) {
                    HMBDC_LOG_C("accept failure, errno=", errno);
                }
                return;
            }
            auto sz = config_.getExt<int>("tcpSendBufferBytes");
            if (sz) {
                if (setsockopt(conn, SOL_SOCKET, SO_SNDBUF, &sz, sizeof(sz)) < 0) {
                    HMBDC_LOG_C("failed to set send buffer size=", sz, " errno=", errno);
                }
            }
            int flag = config_.getExt<bool>("nagling")?0:1;
            if (setsockopt(conn, IPPROTO_TCP, TCP_NODELAY, (char*) &flag, sizeof(flag)) < 0) {
                HMBDC_LOG_C("failed to set TCP_NODELAY, errno=", errno);
            }
            try {
                auto k = advertisingMessage_.connKey;
                auto s = std::make_shared<SendSession>(conn, k, toSendQueue_, maxSendBatch_);
                sessions_.insert(s);
            } catch (std::exception const& e) {
                HMBDC_LOG_C(e.what());
            }
        }
    }
    
    Config const& config_;
    using Sessions = unordered_set<SendSession::ptr>;
    Sessions sessions_;
    ToSendQueue toSendQueue_; 
    EpollFd serverFd_;
    sockaddr_in& serverAddr_;
    mutable TopicSource advertisingMessage_;
    mutable std::default_random_engine generator_;
    mutable std::uniform_int_distribution<uint64_t> distribution_;

    size_t maxSendBatch_;
};
} //sendserver_detail
using SendServer = sendserver_detail::SendServer;
}}}
