#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Logger.hpp"
#include "hmbdc/app/tcpcast/Transport.hpp"
#include "hmbdc/app/tcpcast/Messages.hpp"

#include "hmbdc/text/StringTrieSet.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/pattern/MonoLockFreeBuffer.hpp"
#include "hmbdc/comm/inet/Misc.hpp"

#include <boost/lexical_cast.hpp>
#include <boost/circular_buffer.hpp>

#include <random>
#include <memory>
#include <utility>
#include <iostream>
#include <fcntl.h>
#include <sys/epoll.h>

namespace hmbdc { namespace app { namespace tcpcast {

namespace sendserver_detail {
struct SendServer;
}

using ToSend = std::vector<iovec>;
using ToSendQueue = boost::circular_buffer<std::tuple<std::pair<char const*, char const*> //topic
    , ToSend
    , size_t //total bytes above
    , hmbdc::pattern::MonoLockFreeBuffer::iterator //end iterator
    >
>;

namespace sendsession_detail {
using namespace std;

struct SendSession {
    using ptr = shared_ptr<SendSession>;
    SendSession(int fd
        , uint64_t connKey
        , ToSendQueue & toSendQueue
        , size_t maxSendBatch)
    : connKey_(connKey)
    , readLen_(0)
    , ready_(false) 
    , toSendQueue_(toSendQueue)
    , toSendQueueIndex_(0)
    , msghdr_{0}
    , msghdrRelic_{0}
    , msghdrRelicSize_(0) {
        msghdrRelic_.msg_iov = new iovec[maxSendBatch * 2]; //double for attachment
        auto addrPort = hmbdc::comm::inet::getPeerIpPort(fd);
        id_ = addrPort.first + ":" + std::to_string(addrPort.second);
        auto forRead = dup(fd);
        if (forRead == -1) {
            HMBDC_THROW(std::runtime_error, "dup failed errno=" << errno);
        }

        readFd_.fd = forRead;
        utils::EpollTask::instance().add(EPOLLIN|EPOLLET, readFd_);
        writeFd_.fd = fd;
        utils::EpollTask::instance().add(EPOLLOUT|EPOLLET, writeFd_);
        HMBDC_LOG_N("SendSession started: ", id());
    }

    ~SendSession() {
        delete [] msghdrRelic_.msg_iov;
        HMBDC_LOG_N("SendSession retired: ", id());
    }

    bool runOnce() HMBDC_RESTRICT {
        if (hmbdc_unlikely(!doRead())) return false;
        if (hmbdc_unlikely(writeFd_.isFdReady() && msghdrRelicSize_)) {
            auto l = sendmsg(writeFd_.fd, &msghdrRelic_, MSG_NOSIGNAL|MSG_DONTWAIT);
            if (hmbdc_unlikely(l < 0)) {
                if (!writeFd_.checkErr()) {
                    HMBDC_LOG_C("sendmsg failed errno=", errno);
                    return false;
                }
                return true;
            } 
            msghdrRelicSize_ -= size_t(l);
            if (hmbdc_unlikely(msghdrRelicSize_)) {
                comm::inet::extractRelicTo(msghdrRelic_, msghdrRelic_, l);
                return true;
            } else {
                toSendQueueIndex_++;
            }
        }
        for (; !msghdrRelicSize_ && writeFd_.isFdReady() 
            && toSendQueueIndex_ != toSendQueue_.size();) {
            if (clientSubscriptions_.check(get<0>(toSendQueue_[toSendQueueIndex_]))) {
                msghdr_.msg_iov = &(get<1>(toSendQueue_[toSendQueueIndex_])[0]);
                msghdr_.msg_iovlen = get<1>(toSendQueue_[toSendQueueIndex_]).size();
                auto l = sendmsg(writeFd_.fd, &msghdr_, MSG_NOSIGNAL|MSG_DONTWAIT);
                if (hmbdc_unlikely(l < 0)) {
                    if (!writeFd_.checkErr()) {
                        HMBDC_LOG_C("sendmsg failed errno=", errno);
                        return false;
                    }
                    return true;
                }
                msghdrRelicSize_ = get<2>(toSendQueue_[toSendQueueIndex_]) - size_t(l);
                if (hmbdc_unlikely(msghdrRelicSize_)) {
                    comm::inet::extractRelicTo(msghdrRelic_, msghdr_, l);
                    return true;
                }
            }
            toSendQueueIndex_++;
        }
        return true;
    }

    char const* id() const {
        return id_.c_str();
    }

    bool ready() const {
        return ready_;
    }

private:
    bool 
    doRead() HMBDC_RESTRICT {
        while (readLen_) {
            auto p = find(data_, data_ + readLen_, '\t');
            if (p != data_ + readLen_) {
                *p = '\000';
                string t(data_ + 1);
                if (hmbdc_unlikely(connKey_)) {
                     if (data_[0] != '@' || stoull(t) != connKey_) {
                        HMBDC_LOG_C("invalid connection attempted by ", id_);
                        return false;
                    }
                    connKey_ = 0;
                } else if (t.size() == 0) {
                    ready_ = true;
                }
                else if (data_[0] == '+') {
                    clientSubscriptions_.add(t);
                } else if (data_[0] == '-') {
                    clientSubscriptions_.erase(t);
                } else {
                        HMBDC_LOG_C("voilating protocle by ", id_);
                        return false;
                }
                memmove(data_, p + 1, readLen_ - (p - data_ + 1));
                readLen_ -= p - data_ + 1;
            } else {
                break; //no complete topic request
            }
        }
        if (hmbdc_unlikely(readFd_.isFdReady())) {
            auto l = recv(readFd_.fd, data_ + readLen_, sizeof(data_) - readLen_,  MSG_NOSIGNAL|MSG_DONTWAIT);
            if (hmbdc_unlikely(l < 0)) {
                if (!readFd_.checkErr()) {
                    HMBDC_LOG_C("recv failed errno=", errno);
                    return false;
                }
                return true;
            } else {
                readLen_ += l;
            }
        };
        return true;
    }
    uint64_t connKey_;
    utils::EpollFd writeFd_;
    utils::EpollFd readFd_;
    char data_[1024];
    size_t readLen_;
    string id_;
    bool ready_;

    friend struct hmbdc::app::tcpcast::sendserver_detail::SendServer;
    ToSendQueue& toSendQueue_;
    ToSendQueue::size_type toSendQueueIndex_;
    msghdr msghdr_;
    msghdr msghdrRelic_;
    size_t msghdrRelicSize_;

    text::StringTrieSet clientSubscriptions_;
};
} //sendsession_detail

using SendSession = sendsession_detail::SendSession;

}}}
