#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Logger.hpp"
#include "hmbdc/app/tcpcast/Transport.hpp"
#include "hmbdc/app/tcpcast/SendServer.hpp"
#include "hmbdc/app/tcpcast/Messages.hpp"
#include "hmbdc/app/udpcast/SendTransportEngine.hpp"
#include "hmbdc//Traits.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/time/Rater.hpp"
#include "hmbdc/numeric/BitMath.hpp"

#include <boost/circular_buffer.hpp>
#include <memory>
#include <tuple>
#include <type_traits>

namespace hmbdc { namespace app { namespace tcpcast {

namespace send_detail {
using namespace hmbdc::time;
using namespace std;
using pattern::MonoLockFreeBuffer;
/**
 * @brief capture the transportation mechanism
 * 
 */
struct SendTransport 
: Transport {
    using ptr = std::shared_ptr<SendTransport>;
    /**
     * @brief ctor
     * @param cfg jason specifing the transport - see example, perf-tcpcast.cpp
     * @param maxMessageSize max messafe size in bytes to be sent
     */
    SendTransport(Config const&, size_t);
    template <typename... Messages>
    void queue(Topic const& t, Messages&&... msgs) {
        auto n = sizeof...(msgs);
        auto it = buffer_.claim(n);
        queue(it, t, std::forward<Messages>(msgs)...);
        buffer_.commit(it, n);
    }

    template <typename... Messages>
    bool tryQueue(Topic const& t, Messages&&... msgs) {
        auto n = sizeof...(msgs);
        auto it = buffer_.tryClaim(n);
        if (it) {
            queue(it, t, std::forward<Messages>(msgs)...);
            buffer_.commit(it, n);
            return true;
        }
        return false;
    }

    template <typename Message, typename ... Args>
    void queueInPlace(Topic const& t, Args&&... args) {
        static_assert(std::is_trivially_destructible<Message>::value, "cannot send message with dtor");
        static_assert(!is_base_of<hasMemoryAttachment, Message>::value 
            || is_first_base_of<hasMemoryAttachment, Message>::value
            , "hasMemoryAttachment has to the first base for Message");
        auto s = buffer_.claim();
        char* addr = static_cast<char*>(*s);
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        h->flag = calculateFlag<Message>();
        h->messagePayloadLen = sizeof(MessageWrap<Message>);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        if (hmbdc_likely(sizeof(Message) <= maxMessageSize_)) {
            new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<Message>(std::forward<Args>(args)...);
        } else {
            HMBDC_THROW(std::out_of_range
                , "maxMessageSize too small to hold a message when constructing SendTransportEngine");
        }
        buffer_.commit(s);
    }

    void queueBytes(Topic const& t, uint16_t tag, void const* bytes, size_t len);

    void stop();

    bool match(Topic const& t) const {
        return boost::regex_match(t.c_str(), topicRegex_);
    }

private:
    std::string topic_;
    boost::regex topicRegex_;

protected:
    size_t maxMessageSize_;
    size_t minRecvToStart_;
    MonoLockFreeBuffer buffer_;

    unique_ptr<udpcast::SendTransport> mcSendTransport_;
    Rater rater_;
    Config mcConfig_;

private:

    template <typename Message>
    static
    uint8_t calculateFlag() {
        if (is_base_of<hasMemoryAttachment, Message>::value) return hasMemoryAttachment::flag;
        return 0;
    }

    template<typename M, typename ... Messages>
    void queue(MonoLockFreeBuffer::iterator it, Topic const& t
        , M&& m, Messages&&... msgs) {
        using Message = typename std::remove_reference<M>::type;
        static_assert(std::is_trivially_destructible<Message>::value, "cannot send message with dtor");
        static_assert(!is_base_of<hasMemoryAttachment, Message>::value 
            || is_first_base_of<hasMemoryAttachment, Message>::value
            , "hasMemoryAttachment has to the first base for Message");

        auto s = *it;
        char* addr = static_cast<char*>(s);
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        h->flag = calculateFlag<Message>();
        h->messagePayloadLen = sizeof(MessageWrap<Message>);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        if (hmbdc_likely(sizeof(Message) <= maxMessageSize_)) {
            new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<Message>(std::forward<M>(m));
        } else {
            HMBDC_THROW(std::out_of_range, "maxMessageSize too small to hold a message when constructing SendTransportEngine");
        }
        queue(++it, t, std::forward<Messages>(msgs)...);
    }

    void queue(MonoLockFreeBuffer::iterator it
        , Topic const& t) {}
};

struct SendTransportEngine
: SendTransport
, TimerManager
, ReoccuringTimer 
, Client<SendTransportEngine> {  
    SendTransportEngine(Config const&, size_t);
    using SendTransport::hmbdcName;
    using SendTransport::schedSpec;

    /*virtual*/ 
    void invokedCb(uint16_t threadSerialNumber) HMBDC_RESTRICT override  {
        runOnce();
        mcSendTransport_->runOnce(true);
    }

    /*virtual*/ 
    bool droppedCb() override {
        stop();
        stopped_ = true;
        return true;
    };

    /**
     * @brief check how many recipient sessions are still active
     * @return numeric_limits<size_t>::max() if the sending hasn't started due to 
     * minRecvToStart has not been met yet
     */
    size_t sessionsRemainingActive() const {
        if (stopped_) return 0;
        if (server_ && !minRecvToStart_) {
            return server_->readySessionCount();
        }
        return numeric_limits<size_t>::max();
    }
private:
    void cullSlow() {
        auto seq = buffer_.readSeq();
        if (seq == lastSeq_ && buffer_.isFull()) {
            server_->killSlowestSession();
        }
        lastSeq_ = seq;
    }

    void runOnce();

    size_t mtu_;
    size_t maxSendBatch_;
    bool waitForSlowReceivers_;
    MonoLockFreeBuffer::iterator begin_;
    ToSend toSend_;
    // SendServer *server_;
    // unique_ptr<SyncSendServer> syncServer_;
    unique_ptr<SendServer> server_;
    size_t lastSeq_;
    bool stopped_;
};
} //send_detail

using SendTransport = send_detail::SendTransport;
using SendTransportEngine = send_detail::SendTransportEngine;
}}}
