#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/udpcast/Transport.hpp"
#include "hmbdc/text/StringTrieSet.hpp"
#include "hmbdc/pattern/MonoLockFreeBuffer.hpp"
#include "hmbdc//Traits.hpp"

#include <boost/lexical_cast.hpp>

#include <memory>
#include <type_traits>

#include <sys/epoll.h>

namespace hmbdc { namespace app { namespace udpcast {

/**
 * @brief interface to power a multicast transport receiving functions
 */
struct RecvTransport : Transport {
    using ptr = std::shared_ptr<RecvTransport>;
    using Transport::Transport;
    virtual ~RecvTransport(){}
    /**
     * @brief a take all arbitrator (no arbitration at all)
     * @details it provides the default type for arbitration which does nothing
     * it also provides a template on writing a user defined arbitrator
     * 
     * @param  h handle to retrieve what is inside of the message
     * @return always returns 1 here
     * In general: 1 keep the message; -1 to drop; 
     * 0 cannot decide, ask me later.
     */
    struct NoOpArb {
        int operator()(TransportMessageHeader const* h) {
            return 1; //always keep it
        }
    };
private:
    friend struct NetContext; //only be used by NetContext
    virtual void listenTo(Topic const& t) = 0;
    virtual void stopListenTo(Topic const& t) = 0;
};

namespace recvtransportengine_detail {
using namespace hmbdc::time;
using namespace std;
using Buffer = hmbdc::pattern::MonoLockFreeBuffer;

/**
 * @class RecvTransportImpl<>
 * @brief impl class
 * 
 * @tparam OutputBuffer type of buffer to hold resulting network messages
 * @tparam MsgArbitrator arbitrator to decide drop or keep messages, suited to arbitrate 
 * between different recv transport. By default, keeping all
 */
template <typename OutputBuffer, typename MsgArbitrator = RecvTransport::NoOpArb>
struct RecvTransportImpl 
: RecvTransport
, MessageHandler<RecvTransportImpl<OutputBuffer, MsgArbitrator>, Subscribe, Unsubscribe> {
    using SELF = RecvTransportImpl;
    using MH = MessageHandler<SELF, Subscribe, Unsubscribe>;
    friend struct NetContext; //only be created by NetContext

/**
 * @brief ctor
 * @details io_service could be passed in by user, in this case NO more than two threads should
 * power this io_service instance since that would violate the thread garantee of Client, which
 * is no callbacks are called in parallel
 * 
 * @param cfg specify the details of the udpcast transport
 * @param outputBuffer holding the results
 * @param arb arbitrator instance to decide which messages to drop and keep; it supports either
 * raw udp packet (BEFORE topic filtering) or hmbdc udpcast message (AFTER topic filtering) level 
 * arbitration depending on which one of 
 * int operator()(void* bytes, size_t len) or 
 * int operator()(TransportMessageHeader const* header) presents in the arb
 */
    RecvTransportImpl(Config const& cfg
        , OutputBuffer& outputBuffer
        , MsgArbitrator arb = NoOpArb())
	: RecvTransport(cfg)
    , buffer_(std::max(sizeof(MessageWrap<Subscribe>), sizeof(MessageWrap<Unsubscribe>))
        , config_.getExt<uint16_t>("cmdBufferSizePower2"))
    , outputBuffer_(outputBuffer)
    , maxItemSize_(outputBuffer.maxItemSize())
    , buf_(new char[mtu_])
    , bufCur_(buf_)
    , bytesRecved_(0)
    , arb_(arb) {
        uint32_t yes = 1;
        if (setsockopt(fd, SOL_SOCKET,SO_REUSEADDR, &yes, sizeof(yes)) < 0) {
            HMBDC_LOG_C("failed to set reuse address errno=", errno);
        }
        auto sz = config_.getExt<int>("udpRecvBufferBytes");
        if (sz) {
            if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &sz, sizeof(sz)) < 0) {
                HMBDC_LOG_C("failed to set send buffer size=", sz);
            }
        }

        auto udpcastListenPort = config_.getExt<uint16_t>("udpcastListenPort");
        struct sockaddr_in udpcastListenAddrPort;
        memset(&udpcastListenAddrPort, 0, sizeof(udpcastListenAddrPort));
        udpcastListenAddrPort.sin_family = AF_INET;
        auto ipStr = cfg.getExt<string>("udpcastListenAddr") == string("ifaceAddr")
            ? comm::inet::getLocalIpMatchMask(cfg.getExt<string>("ifaceAddr")):cfg.getExt<string>("udpcastListenAddr");
        udpcastListenAddrPort.sin_addr.s_addr = inet_addr(ipStr.c_str());
        udpcastListenAddrPort.sin_port = htons(udpcastListenPort);
        if (bind(fd, (struct sockaddr *)&udpcastListenAddrPort, sizeof(udpcastListenAddrPort)) < 0) {
            HMBDC_THROW(runtime_error, "failed to bind unicast udpcast listen address " 
                << ipStr << ':' << cfg.getExt<short>("udpcastListenPort") << " errno=" << errno);
        }

        if ((udpcastListenAddrPort.sin_addr.s_addr & 0x000000F0) == 0xE0) {//multicast address
            /* use setsockopt() to request that the kernel join a multicast group */
            struct ip_mreq mreq;
            mreq.imr_multiaddr.s_addr = udpcastListenAddrPort.sin_addr.s_addr;
            auto iface = 
                comm::inet::getLocalIpMatchMask(config_.getExt<string>("ifaceAddr")); 
            mreq.imr_interface.s_addr=inet_addr(iface.c_str());
            if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0) {
                HMBDC_THROW(runtime_error, "failed to join " << ipStr << ':'
                    << cfg.getExt<short>("udpcastPort"));
            }
        }
    }

    ~RecvTransportImpl() {
        delete [] buf_;
    }

/**
 * @brief start the show by schedule the mesage recv
 */
    void start() {
        utils::EpollTask::instance().add(EPOLLIN|EPOLLET, *this);
    }

    void runOnce(bool alwaysPoll = false) HMBDC_RESTRICT {
        typename Buffer::iterator begin, end;
        auto n = buffer_.peek(begin, end);
        auto it = begin;
        while (it != end) {
            MH::handleMessage(*static_cast<MessageHead*>(*it++));
        }
        buffer_.wasteAfterPeek(begin, n);
        if (hmbdc_unlikely(alwaysPoll || !isFdReady())) {
            utils::EpollTask::instance().poll();
        }
        resumeRead();
    }
   
/**
 * @brief only used by MH
 */
    void handleMessageCb(Subscribe const& t) {
        subscriptions_.add(boost::lexical_cast<std::string>(t.topic));
    }

/**
 * @brief only used by MH
 */
    void handleMessageCb(Unsubscribe const& t) {
        subscriptions_.erase(boost::lexical_cast<std::string>(t.topic));
    }


    void listenTo(Topic const& t) override {
        buffer_.put(MessageWrap<Subscribe>{t});
    }

    void stopListenTo(Topic const& t) override {
        buffer_.put(MessageWrap<Unsubscribe>{t});
    }
    sockaddr_in udpcastListenRemoteAddr = {0};
private:
    template <bool is_raw_arb>
    typename enable_if<is_raw_arb, int>::type applyRawArb(void* pkt, size_t len) {
        return arb_(pkt, len);
    }

    template <bool is_raw_arb>
    typename enable_if<!is_raw_arb, int>::type applyRawArb(void*, size_t) {
        return 1;
    }

    template <bool is_raw_arb>
    typename enable_if<!is_raw_arb, int>::type applyArb(TransportMessageHeader* h) {
        return arb_(h);
    }

    template <bool is_raw_arb>
    typename enable_if<is_raw_arb, int>::type applyArb(TransportMessageHeader* h) {
        return 1;
    }

    void resumeRead() HMBDC_RESTRICT {
        using traits = 
            function_traits<typename std::remove_reference<MsgArbitrator>::type>;
        using arg0 = typename traits::template arg<0>::type;
        bool const is_raw_arb = std::is_same<void*, typename std::decay<arg0>::type>::value;
        do {
            if (bytesRecved_ && bufCur_ == buf_) {
                auto a = applyRawArb<is_raw_arb>(buf_, bytesRecved_);
                if (hmbdc_unlikely(a == 0)) {
                    //keep bufCur_ and bytesRecved_ unchanged
                    return;
                } else if (hmbdc_unlikely(a < 0)) { //drop the whole packet
                    bytesRecved_ = 0;
                } // else process it
            }
            while (bytesRecved_) {
                auto h = reinterpret_cast<TransportMessageHeader*>(bufCur_);
                auto wireSize = h->wireSize();

                if (hmbdc_likely(bytesRecved_ >= wireSize)) {
                    if (subscriptions_.check(h->topic())) {
                        if (hmbdc_unlikely(wireSize > bytesRecved_)) {
                            break;
                        }
                        auto a = applyArb<is_raw_arb>(h);

                        if (hmbdc_likely(a > 0)) {
                            auto l = std::min<size_t>(maxItemSize_, h->messagePayloadLen());
                            outputBuffer_.put(h->payload(), l);
                        } else if (a == 0) {
                            //keep bufCur_ and bytesRecved_ unchanged
                            return;
                        } //else drop and move on
                    }
                    bytesRecved_ -= wireSize;
                    bufCur_ += wireSize;
                } else {
                    break;
                }
            }
            bytesRecved_ = 0;
            bufCur_ = buf_;
            if (isFdReady()) {
                socklen_t addrLen = sizeof(udpcastListenRemoteAddr);
                auto l = recvfrom(fd, buf_, mtu_, MSG_NOSIGNAL|MSG_DONTWAIT
                    , (struct sockaddr *)&udpcastListenRemoteAddr, &addrLen);
                if (hmbdc_unlikely(l < 0)) {
                    if (!checkErr()) {
                        HMBDC_LOG_C("recvmmsg failed errno=", errno);
                    }
                    return;
                } else if (l == 0) {
                    //nothing to do now
                    return;
                }
                bytesRecved_ = l;
            }
        } while(bytesRecved_);
    }

    Buffer buffer_;
    OutputBuffer& outputBuffer_;
    size_t maxItemSize_;
    char* buf_;
    char* bufCur_;
    size_t bytesRecved_;
    text::StringTrieSet subscriptions_;
    MsgArbitrator arb_;
};

template <typename OutputBuffer, typename MsgArbitrator = RecvTransport::NoOpArb>
struct RecvTransportEngineImpl 
: RecvTransportImpl<OutputBuffer, MsgArbitrator>
, Client<RecvTransportEngineImpl<OutputBuffer, MsgArbitrator>> {
    using RecvTransportImpl<OutputBuffer, MsgArbitrator>::RecvTransportImpl;
    using RecvTransportImpl<OutputBuffer, MsgArbitrator>::hmbdcName;
    using RecvTransportImpl<OutputBuffer, MsgArbitrator>::schedSpec;

/**
 * @brief power the io_service and other things
 * 
 */
    /*virtual*/ 
    void invokedCb(uint16_t) HMBDC_RESTRICT override  {
        RecvTransportImpl<OutputBuffer, MsgArbitrator>::runOnce();
    }


/**
 * @brief start the show by schedule the mesage recv
 */
    /*virtual*/ 
    void messageDispatchingStartedCb(uint16_t threadSerialNumber) override {
        RecvTransportImpl<OutputBuffer, MsgArbitrator>::start();
    };
    
/**
 * @brief should not happen ever unless an exception thrown
 * 
 * @param e exception thown
 */
    /*virtual*/ 
    void stoppedCb(std::exception const& e) override {
        HMBDC_LOG_C(e.what());
    };

};

} //recvtransportengine_detail

template <typename OutputBuffer, typename MsgArbitrator = RecvTransport::NoOpArb>
using RecvTransportImpl = recvtransportengine_detail::RecvTransportImpl<OutputBuffer, MsgArbitrator>;

template <typename OutputBuffer, typename MsgArbitrator = RecvTransport::NoOpArb>
using RecvTransportEngine = recvtransportengine_detail::RecvTransportEngineImpl<OutputBuffer, MsgArbitrator>;
 
}}}
