#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/udpcast/Transport.hpp"
#include "hmbdc/app/udpcast/Messages.hpp"
#include "hmbdc/app/Base.hpp"
#include "hmbdc/comm/inet/Endpoint.hpp"
#include "hmbdc/pattern/MonoLockFreeBuffer.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/time/Rater.hpp"
#include "hmbdc/numeric/BitMath.hpp"

#include <boost/regex.hpp>

#include <memory>
#include <iostream>

#include <sys/epoll.h>
namespace hmbdc { namespace app { namespace udpcast {

namespace sendtransportengine_detail {
using namespace std;    
using namespace hmbdc::time;
using namespace hmbdc::pattern;
using Buffer = hmbdc::pattern::MonoLockFreeBuffer;

struct SendTransport 
: Transport {
    using ptr = std::shared_ptr<SendTransport>;
    SendTransport(Config const&, size_t); 
    ~SendTransport();

    bool match(Topic const& t) const {
        return boost::regex_match(t.c_str(), topicRegex_);
    }

    template <typename... Messages>
    void queue(Topic const& t, Messages&&... msgs) {
        auto n = sizeof...(msgs);
        auto it = buffer_.claim(n);
        queue(it, t, std::forward<Messages>(msgs)...);
        buffer_.commit(it, n);
    }

    template <typename... Messages>
    bool tryQueue(Topic const& t, Messages&&... msgs) {
        auto n = sizeof...(msgs);
        auto it = buffer_.tryClaim(n);
        if (it) {
            queue(it, t, std::forward<Messages>(msgs)...);
            buffer_.commit(it, n);
            return true;
        }
        return false;
    }

    template <typename Message, typename ... Args>
    void queueInPlace(Topic const& t, Args&&... args) {
        static_assert(std::is_trivially_destructible<Message>::value, "cannot send message with dtor");
        auto s = buffer_.claim();
        char* addr = static_cast<char*>(*s);
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        if (hmbdc_likely(sizeof(Message) <= maxMessageSize_)) {
            new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<Message>(std::forward<Args>(args)...);
            h->messagePayloadLen() = sizeof(MessageWrap<Message>);
        } else {
            HMBDC_THROW(std::out_of_range
                , "maxMessageSize too small to hold a message when constructing SendTransportEngine");
        }
        buffer_.commit(s);
    }

    void queueBytes(Topic const& t, uint16_t tag, void const* bytes, size_t len);

    void runOnce(bool alwaysPoll = false) HMBDC_RESTRICT {
        resumeSend();
        if (hmbdc_unlikely(alwaysPoll || !isFdReady())) {
            utils::EpollTask::instance().poll();
        }
    }
    
    template <typename M>
    static TransportMessageHeader* encode(Topic const& t, void* buf, size_t bufLen, M const& m) {
        using Message = typename std::remove_reference<M>::type;
        char* addr = static_cast<char*>(buf);
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        if (hmbdc_likely(sizeof(Message) <= bufLen)) {
            new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<Message>(m);
            h->messagePayloadLen() = sizeof(MessageWrap<Message>);
        } else {
            HMBDC_THROW(std::out_of_range, "bufLen too small to hold a message");
        }
        return h;
    }

    void stop();

private:
    std::string topic_;
    boost::regex topicRegex_;  
    size_t maxMessageSize_;
    typename Buffer::iterator begin_, it_, end_;
    Buffer buffer_;
    Rater rater_;
    size_t maxSendBatch_;

    iovec* toSendMsgs_;
    size_t toSendMsgsHead_;
    size_t toSendMsgsTail_;
    mmsghdr* toSendPkts_;
    size_t toSendPktsHead_;
    size_t toSendPktsTail_;
    std::vector<comm::inet::Endpoint> udpcastDests_;

    uint16_t
    outBufferSizePower2();

    template<typename M, typename ... Messages>
    void queue(typename Buffer::iterator it, Topic const& t
        , M&& m, Messages&&... msgs) {
        using Message = typename std::remove_reference<M>::type;
        static_assert(std::is_trivially_destructible<Message>::value, "cannot send message with dtor");
        auto s = *it;
        char* addr = static_cast<char*>(s);
        auto h = reinterpret_cast<TransportMessageHeader*>(addr);
        auto tl = t.copyTo(addr + sizeof(TransportMessageHeader));
        h->topicLen = tl;
        if (hmbdc_likely(sizeof(Message) <= maxMessageSize_)) {
            new (addr + sizeof(TransportMessageHeader) + tl) MessageWrap<Message>(std::forward<M>(m));
            h->messagePayloadLen() = sizeof(MessageWrap<Message>);
        } else {
            HMBDC_THROW(std::out_of_range
                , "maxMessageSize too small to hold a message when constructing SendTransportEngine");
        }
        queue(++it, t, std::forward<Messages>(msgs)...);
    }

    void queue(typename Buffer::iterator it
        , Topic const& t) {} 
    void resumeSend();
};

struct SendTransportEngine
: SendTransport
, Client<SendTransportEngine> {  
    using SendTransport::SendTransport;
    using SendTransport::hmbdcName;
    using SendTransport::schedSpec;

    /*virtual*/ 
    void invokedCb(uint16_t) HMBDC_RESTRICT override  {
        runOnce();
    }
    /*virtual*/ bool droppedCb() override {
        stop();
        return true;
    };
};

} //sendtransportengine_detail
using SendTransport = sendtransportengine_detail::SendTransport;
using SendTransportEngine = sendtransportengine_detail::SendTransportEngine;
}}}
