#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Base.hpp"
#include "hmbdc/text/Misc.h"
#include "hmbdc/numeric/BitMath.hpp"
#include "hmbdc/os/Signals.hpp"

#include <iostream>
#include <vector>
#include <memory>
#include <unistd.h>
#include <boost/program_options.hpp>

extern const char *__progname_full;


/**
 * @namespace hmbdc::app::utils
 * utilities based on app and serve its higher layers
 */

namespace hmbdc { namespace app { namespace utils {

namespace netperf_detail {
using namespace tested_module;
using namespace hmbdc::time;
using namespace hmbdc::numeric;
using namespace boost::property_tree;
using namespace std;

using SendCtx = hmbdc::app::Context<24>;

struct Message
: hasTag<1001> {
    Message(uint32_t len) 
    : len(len)
    , seq(0)
    {}
    uint32_t len;
    uint32_t seq;
    friend std::ostream& operator << (std::ostream& os, Message const& r) {
        os << "Message " << ' ' << r.len << ' ' << r.seq;
        return os;
    }
    char padding[65536 - sizeof(seq) - sizeof(len)];
} __attribute__((__packed__));

struct MessageAtt
: hasMemoryAttachment
, hasTag<1002> {
    MessageAtt() 
    : seq(0)
    {}
    uint32_t seq;
    friend std::ostream& operator << (std::ostream& os, MessageAtt const& r) {
        os << "MessageAtt " << ' ' << r.len << ' ' << r.seq;
        return os;
    }
};

struct PeriodicReporter : TimerManager, ReoccuringTimer {
    PeriodicReporter()
    : ReoccuringTimer(Duration::seconds(1u)) {
        setCallback(
            [this](TimerManager& tm, SysTime const& now) {
            report();
            }
        );
        schedule(SysTime::now(), *this);
    }

    virtual void report() = 0; 
};

struct SenderClient 
: Client<SenderClient>
, PeriodicReporter {
    SenderClient(SendCtx& ctx, Topic const topic, size_t msgSize, bool allowAtt)
    : ctx_(ctx)
    , allowAtt_(allowAtt)
    , msg_(msgSize)
    , msgAtt_()
    , periodicGoodMsgCount_(0) 
    , periodStartAt_(SysTime::now()) {
        sender_ = NetContext::instance().getSender(topic);
        if (msg_.len > 1000 && allowAtt) {
            fileMap_.map(__progname_full);
            if (fileMap_.len < msgSize) {
                HMBDC_THROW(std::out_of_range, "exceeds test allowed msg size <=" << fileMap_.len);
            }
            msgAtt_.attachment = fileMap_.attachment;
            msgAtt_.len = msgSize;
            //don't release memory, since we r reusing the same msfAtt_ everytime
            msgAtt_.afterConsumedCleanupFunc = nullptr;
        }
    }

    ~SenderClient() {
        if (msgAtt_.len > 1000 && allowAtt_) fileMap_.unmap();
    }
    char const* hmbdcName() const { 
        return "perf-tx"; 
    }
    
    /*virtual*/ 
    void invokedCb(uint16_t) HMBDC_RESTRICT override {
        //send as fast as possible since the engine has rate control
        if (hmbdc_likely(!allowAtt_ || msg_.len <= 1000u)) {
            sender_->send(msg_, msg_.len);
            msg_.seq++;
            periodicGoodMsgCount_++;
        } else {
            sender_->send(msgAtt_);
            msgAtt_.seq++;
            periodicGoodMsgCount_++;
        }
    }

    void stoppedCb(std::exception const& e) override {
        std::cout << (e.what()) << std::endl;
    };

    void report() override {
        auto rate = size_t(periodicGoodMsgCount_ / ((SysTime::now() - periodStartAt_) / Duration::seconds(1)));
        auto l = msg_.len;
        HMBDC_LOG_n("msgSize=", l, " mps=", rate);
        periodicGoodMsgCount_ = 0u;
        periodStartAt_ = SysTime::now();
    }

    SendCtx& ctx_;
    bool allowAtt_;
    Sender* sender_;
    Message msg_;
    MessageAtt msgAtt_;
    size_t periodicGoodMsgCount_;
    SysTime periodStartAt_;
    hasMemoryAttachment fileMap_;
};

void runInSenderMode(Config & config, Topic const& topic, size_t msgSize
    , vector<uint16_t> cpus, bool allowAtt) {
    using namespace std;
    SendCtx ctx; 
    SingletonGuardian<NetContext> g;
    auto engine = NetContext::instance().createSendTransportEngine(config
        , msgSize > 1000 && allowAtt
            ?std::max(sizeof(MessageAtt), sizeof(StartMemorySegTrain))
            :msgSize);
    SenderClient client(ctx, topic, msgSize, allowAtt);
    ctx.start(*engine, 1ul << cpus[0]
        , client, 1ul << cpus[1]);
    os::HandleSignals::onTermIntDo(
        [&ctx]() {
            ctx.stop();
        }
    );
    ctx.join();
}

using RecvCtx = Context<0, context_property::broadcast<1>>;
struct ReceiverClient 
: Client<ReceiverClient, Message, MessageAtt>
, PeriodicReporter {
    ReceiverClient(RecvCtx& ctx, size_t msgOverhead)
    : ctx_(ctx)
    , periodicGoodMsgCount_(0)
    , periodicGoodMsgBytes_(0)
    , periodicMsgCount_(0)
    , periodicMsgBytes_(0)
    , periodStartAt_(SysTime::now())
    , seq_(0)
    , periodicMissed_(0)
    , periodicOoo_(0)
    , periodicDup_(0)
    , msgSize_(0)
    , msgOverhead_(msgOverhead) {
    }

    char const* hmbdcName() const { 
        return "perf-rx"; 
    }
    
    template<typename M>
    void handleMessageCb(M const& r) {
        msgSize_ = r.len;
        if (hmbdc_likely(r.seq > seq_)) {
            periodicGoodMsgCount_++;
            periodicGoodMsgBytes_ += msgOverhead_ + msgSize_;
            if (hmbdc_likely(seq_)) periodicMissed_ += r.seq - seq_ - 1;
        } else if (r.seq == seq_) {
            periodicDup_++;
        } else {
            periodicOoo_++;
        }
        seq_ = r.seq;
        periodicMsgCount_++;
        periodicMsgBytes_ += msgOverhead_ + msgSize_;
    }

    void handleMessageCb(MessageAtt& r) {
        handleMessageCb<MessageAtt>(r);
        r.release();
    }

    void stoppedCb(std::exception const& e) override {
        std::cout << (e.what()) << std::endl;
    };

    void report() override {
        auto rateGood = size_t(periodicGoodMsgCount_ / ((SysTime::now() - periodStartAt_) / Duration::seconds(1)));
        HMBDC_LOG_n("msgSize=", msgSize_, " msgOverhead=", msgOverhead_, " mps(good)=", rateGood, " Mbps(good)=", periodicGoodMsgBytes_*8u >> 20u
            , " missed=" , periodicMissed_ , '(' , periodicMissed_?periodicMissed_ * 100ul / (periodicMissed_ + periodicMsgCount_):0
            , "%) outOfOrder=" , periodicOoo_ 
            , " duplicate=" , periodicDup_ 
            , " Mbps(all)=" , periodicMsgBytes_*8u >> 20u
        );

        periodicGoodMsgCount_ = 0u;
        periodicGoodMsgBytes_ = 0u;
        periodicMsgCount_ = 0;
        periodicMsgBytes_ = 0u;
        periodicMissed_ = 0;
        periodicDup_ = 0;
        periodicOoo_ = 0;

        periodStartAt_ = SysTime::now();
    }

    RecvCtx& ctx_;
    size_t periodicGoodMsgCount_;
    size_t periodicGoodMsgBytes_;
    size_t periodicMsgCount_;
    size_t periodicMsgBytes_;
    SysTime periodStartAt_;
    size_t seq_;
    size_t periodicMissed_;
    size_t periodicOoo_;
    size_t periodicDup_;
    size_t msgSize_;
    size_t msgOverhead_;
};

template <typename RecvCtorArgsTuple>
void runInReceiverMode(Config const& config, Topic const& topic, size_t msgSize, uint16_t inBuferSizePower2
    , vector<uint16_t> cpus, RecvCtorArgsTuple&& recvCtorArgs, bool allowAtt) {
    using namespace std;
    using namespace boost;
    RecvCtx ctx(inBuferSizePower2, 1, msgSize > 1000 && allowAtt?sizeof(MessageAtt):msgSize);   

    SingletonGuardian<NetContext> g;
    auto& net = NetContext::instance();
    
    auto engine = net.createRecvTransportEngineTuply(config, ctx.buffer(), std::forward<RecvCtorArgsTuple>(recvCtorArgs));
    auto msgOverhead = 8u + sizeof(TransportMessageHeader) + topic.size();
    ReceiverClient client(ctx, msgOverhead);
    net.listenTo(topic);
    ctx.start(*engine, 1ul << cpus[0]
        , client, 1ul << cpus[1]);
    os::HandleSignals::onTermIntDo(
        [&ctx]() {
            ctx.stop();
        }
    );
    ctx.join();

    net.stopListenTo(topic);
}

template <typename RecvCtorArgsTuple>
int startNetPerf(char const* dftUserConfig, char const* helpStrIn, int argc, char** argv
    , RecvCtorArgsTuple&& recvCtorArgs, bool allowAtt = true) {
    SingletonGuardian<SyncLogger> logGuard(cout);
    using namespace std;
    string role;
    string userCfg;
    size_t msgSize;
    vector<uint16_t> cpus;
    uint16_t inBuferSizePower2;
    Topic topic("t");
	if (argc > 1) role = string(argv[1]);
    if (role != string("--sender")
		&& role != string("--receiver")) {
			cerr << "1st arg needs to be either --sender or --receiver" << endl;
		exit(1);
	}
	argc--;
	argv++;

    bool isSender = role == string("--sender");
    auto checkMsgSize = [](size_t s) {
        if (s < 8) HMBDC_THROW(out_of_range, "msgSize too small (>=8)");
    };
    namespace po = boost::program_options;
    po::options_description desc("additional allowed options (after --sender or --receiver)");
    string helpStr =
    "This program can be used to test messaging performance among a group of hosts.\n"
    "It can run as either a sender or a receiver. Normally one sender for a single interface in a test. \n"
    "When Sender sends out traffic, receivers subscribe to the traffic and collect stats.\n"
    "ctrl-c terminate.\n";
    helpStr += string(helpStrIn);
    desc.add_options()
    ("help", helpStr.c_str())
    ("cfg,c", po::value<std::string>(&userCfg), "use the user provided config file and ignore the ones in command options")
    ("print", "print effective cfg values")
    ("cpus", po::value(&cpus)->multitoken()->default_value({0, 1}, "0 1"), "specify the 2 cpu index numbers used in the test (1 for engine and 1 for msg client)")
    ("msgSize", po::value<size_t>(&msgSize)->default_value(8u)->notifier(checkMsgSize), "size of the message (8-size of this executable file). when > 1000 the message is sent using zero copy memory attachment if it is supported by the transport type and hmbdc posts no limit on attachment size")
    ("inBuferSizePower2", po::value<uint16_t>(&inBuferSizePower2)->default_value(0u), "2^inBufferSizePower2 is the number of message that can be buffered in the program, default 0 means automatically caulcated based on 32MB as lower bound")
    ("topic", po::value<Topic>(&topic)->default_value(topic), "topic used in the test");

    Config dft(dftUserConfig, isSender?"tx":"rx");
    auto params = dft.content();
    for (auto it = params.begin(); it != params.end();) {
    	string& name = it->first;
    	string& val = it->second;
    	it++;
    	string& comment = it->second;
    	it++;
    	desc.add_options()
    		(name.c_str(), po::value<string>(&val)->default_value(val), comment.c_str());
    }
    
    po::positional_options_description p;
    po::variables_map vm;
    po::store(po::command_line_parser(argc, argv).
        options(desc).positional(p).run(), vm);
    po::notify(vm);

    if (argc == 1 || vm.count("help") || msgSize < 8u || cpus.size() != 2) {
        cout << desc << "\n";
        return 0;
    }

    Config config;
    if (userCfg.size()) {
        std::istringstream cs(userCfg);
        config = Config(cs, isSender?"tx":"rx");
    } else {
        for (auto& p : params) {
        	config.put(p.first, p.second);
        }
    }
    if (msgSize > 1000 && allowAtt && !isSender) {
        config.put("allowRecvMemoryAttachment", "1002");
    }

    if (vm.count("print")) {
        write_json(cout, config);
        exit(0);
    }

    if (isSender) {
        runInSenderMode(config, topic, msgSize, cpus, allowAtt);
    } else {
        if (!inBuferSizePower2) {
            inBuferSizePower2 = hmbdc::numeric::log2Upper(32ul * 1024ul * 1024ul / (8ul + msgSize));
            cout << "auto set --inBuferSizePower2=" << inBuferSizePower2 << endl;
        }
        runInReceiverMode(config, topic, msgSize, inBuferSizePower2
            , cpus, forward<RecvCtorArgsTuple>(recvCtorArgs), allowAtt);
    }
    return 0;
}
} //netperf_detail

using netperf_detail::startNetPerf;
}}}

