#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Base.hpp"
#include "hmbdc/numeric/Stat.hpp"
#include "hmbdc/time/Rater.hpp"
#include "hmbdc/os/Signals.hpp"

#include <iostream>
#include <memory>
#include <functional>
#include <unistd.h>
#include <sys/mman.h>

namespace hmbdc { namespace app { namespace utils {

namespace pingpong_detail {
using namespace hmbdc::time;
using namespace hmbdc::numeric;
using namespace std;

struct Ball
: hasTag<1001> {
    Ball(size_t s) 
    : size(s)
    , ts(SysTime::now())
    {}
    size_t size;
    SysTime ts;
    char padding[1013 - 16];
};

template <typename Sender>
struct Pinger 
: Client<Pinger<Sender>, Ball> {
    Pinger(Sender* sender, uint16_t msgPerSec, uint16_t msgSize, size_t skipFirst)
    : rater_(Duration::seconds(1), msgPerSec, 1u)
    , sender_(sender)
    , periodicPingCount_(0)
    , pingCount_(0)
    , skipped_(skipFirst)
    , msgPerSec_(msgPerSec)
    , msgSize_(msgSize) {
    }

    char const* hmbdcName() const { 
        return "pinger"; 
    } 

    void messageDispatchingStartedCb(uint16_t threadSerialNumber) override {
        cout << "Started with the first " << skipped_ << " values ignored(x), press ctrl-c to get results" << endl;
    };

    void handleMessageCb(Ball const& m) {
        auto now = SysTime::now();
        auto lat = now - m.ts;
        if (!skipped_) { 
            stat_.add(lat);
        } else {
            --skipped_;
        }
    }

    void stoppedCb(exception const& e) override {
        cerr << e.what() << endl;
    };

    void invokedCb(uint16_t) override {
        if (hmbdc_unlikely(rater_.check())) {
            if (++periodicPingCount_ == msgPerSec_) {
                cout << (skipped_?'x':'.') << flush;
                periodicPingCount_ = 0;
            }
            Ball p(msgSize_);
            sender_->send(p, msgSize_);
            rater_.commit();
            if (!skipped_) pingCount_++;
        }
    }

    void finalReport() {
        cout << "\nround trip time (sec):(" << stat_.sampleSize() << '/' << pingCount_ << "):";
        cout << stat_;
        cout << endl;
    }

    Rater rater_;
    Sender* sender_;
    Stat<Duration> stat_;
    size_t periodicPingCount_;
    size_t pingCount_;
    size_t skipped_;
    uint16_t msgPerSec_;
    uint16_t msgSize_;
};


template <typename Sender>
struct Ponger 
: Client<Ponger<Sender>, Ball> {
    Ponger(Sender* sender)
    : sender_(sender) {
    }

    char const* hmbdcName() const { 
        return "ponger"; 
    }

    void messageDispatchingStartedCb(uint16_t) override {
        cout << "Started, press ctrl-c to stop" << endl;
    };

    void handleMessageCb(Ball const& m) {
        sender_->send(m, m.size);
    }

    void stoppedCb(exception const& e) override {
        cerr << e.what() << endl;
    };

    Sender* sender_;
};

template <typename NetContext>
int 
pingpong(Config const& config, vector<uint16_t> cpus) {
    size_t skipFirst = config.getExt<size_t>("skipFirst");
    bool ping = config.getExt<bool>("ping");
    auto runTime = config.getExt<uint32_t>("runTime");

    SingletonGuardian<NetContext> g;
    auto& net = NetContext::instance(); 
    auto msgSize = config.getExt<uint16_t>("msgSize");
    msgSize = min<uint16_t>(msgSize, 512);
    msgSize = max<uint16_t>(msgSize, 16);
    using MyContext = Context<>;
    MyContext ctx(14
        , 10
        , msgSize
    );
    Config pingCfg(config, "ping");
    Config pongCfg(config, "pong");
    cout << "initailizing..." << endl;
    if (ping) {
        auto sengine = net.createSendTransportEngine(pingCfg, msgSize);
        auto rengine = net.createRecvTransportEngine(pongCfg, ctx.buffer());
        Pinger<typename NetContext::Sender> pinger(net.getSender("ping")
            , config.getExt<uint16_t>("msgPerSec")
            , msgSize
            , skipFirst
        );
        net.listenTo("pong");

        ctx.start(*rengine, 1ul << cpus[0]
            , *sengine, 1ul << cpus[1]
            , pinger,   1ul << cpus[2]);
        os::HandleSignals::onTermIntDo(
            [&ctx]() {
            ctx.stop();
        });
        if (runTime) {
            sleep(runTime);
            ctx.stop();
        }

        ctx.join();

        pinger.finalReport();
    } else {
        auto sengine = net.createSendTransportEngine(pongCfg, msgSize);
        auto rengine = net.createRecvTransportEngine(pingCfg, ctx.buffer());

        Ponger<typename NetContext::Sender> ponger(net.getSender("pong"));
        net.listenTo("ping");
        ctx.start(*rengine, 1ul << cpus[0]
            , *sengine, 1ul << cpus[1]
            , ponger,   1ul << cpus[2]);
        os::HandleSignals::onTermIntDo(
            [&ctx]() {
            ctx.stop();
        });
        if (runTime) {
            sleep(runTime);
            ctx.stop();
        }
    
        ctx.join();
    }
    return 0;
}

} //pingpong_detail

using pingpong_detail::pingpong;
}}}
