#include "hmbdc/Copyright.hpp"
#pragma once 

#include "hmbdc/Exception.hpp"
#include <stdexcept>
#include <type_traits>
#include <typeinfo>

namespace hmbdc { namespace pattern {
/**
 * @class SingletonGuardian<>
 * @brief RAII representing the lifespan of the underlying Singleton 
 * which also ganrantees the singularity of underlying Singleton
 * @details when the SingletonGuardian is constructored, the underlying Singleton is created;
 * when the SingletonGuardian goes out of scope the dtor of the Singleton is called.
 * 
 * @tparam Singleton the underlying type, which needs to be derived from GuardedSingleton
 */
template <typename Singleton>
struct SingletonGuardian {
    template <typename...Args>
    SingletonGuardian(Args&&...);
    virtual ~SingletonGuardian();
};

/**
 * @class GuardedSingleton<>
 * @brief base for the Singleton that works with SingletonGuardian 
 * @details a good practice is to declare the ctor of the derived class private
 * and friend the derived with the SingletonGuardian<derived>
 * 
 * @tparam Singleton the derived type
 */
template<typename Singleton>
struct GuardedSingleton {
    friend struct SingletonGuardian<Singleton>;
    static Singleton& instance() {return *pInstance_s;}
    static bool initialized() {return pInstance_s;}
    using element_type = Singleton;

protected:
    GuardedSingleton() = default;

private:
    static Singleton* pInstance_s;
};

template <typename Singleton> Singleton*
    GuardedSingleton<Singleton>::pInstance_s = 0;

template <typename Singleton>
template <typename...Args>
SingletonGuardian<Singleton>::
SingletonGuardian(Args&&...args) {
    if (GuardedSingleton<Singleton>::pInstance_s) {
        HMBDC_THROW(std::runtime_error
            , "Cannot reinitialize typeid=" << typeid(Singleton).name());
    }
    GuardedSingleton<Singleton>::pInstance_s 
        = new Singleton(std::forward<Args>(args)...);
}

template <typename Singleton>
SingletonGuardian<Singleton>::
~SingletonGuardian() {
    delete GuardedSingleton<Singleton>::pInstance_s;
    GuardedSingleton<Singleton>::pInstance_s = 0;
}

}}


